#ifndef TX_COMMON_H
#define TX_COMMON_H

#include <cstddef>
#include <stdio.h>
#include <iostream>
#include <algorithm> // std::min
#include <cuda.h>
#include <cuComplex.h>
#ifdef CUSP_USE_TEXTURE_MEMORY
#include <cusp/detail/device/texture.h>
#endif
#include <cusp/detail/device/arch.h>
#include <cusp/detail/device/common.h>
#include <cusp/detail/device/utils.h>
#include <cusp/version.h>

#include <cusparse_v2.h>

static inline char * cusparseGetErrorString(cusparseStatus_t stat) {
  switch (stat) {
  case CUSPARSE_STATUS_NOT_INITIALIZED:
    return const_cast<char *>(" CuSPARSE not initialized");
  case CUSPARSE_STATUS_ALLOC_FAILED:
    return const_cast<char *>(" CuSPARSE allocation failed");
  case CUSPARSE_STATUS_INVALID_VALUE:
    return const_cast<char *>(" CuSPARSE invalid input argument");
  case CUSPARSE_STATUS_ARCH_MISMATCH:
    return const_cast<char *>(" CuSPARSE architecture mismatch");
  case CUSPARSE_STATUS_MAPPING_ERROR:
    return const_cast<char *>(" CuSPARSE mapping error");
  case CUSPARSE_STATUS_EXECUTION_FAILED:
    return const_cast<char *>(" CuSPARSE execution failed");
  case CUSPARSE_STATUS_INTERNAL_ERROR:
    return const_cast<char *>(" CuSPARSE internal error");
  case CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED:
    return const_cast<char *>(" CuSPARSE matrix type not supported");
  default:
    return const_cast<char *>(" CuSPARSE success");
  }  
}

#define CUSPARSE_ERR_CHK(stat, tx_function, cusparse_function) if (stat != CUSPARSE_STATUS_SUCCESS) {std::cout<<"Failed in "<<tx_function<<" ... "<<cusparse_function<<" : "<<cusparseGetErrorString(stat)<<std::endl; return stat; }
#define CUDA_ERR_CHK(err, tx_function, cuda_function) if (err != cudaSuccess) {std::cout<<"Failed in "<<tx_function<<" ... "<<cuda_function<<" ... cuda error "<<cudaGetErrorString(err) <<std::endl; return err; }

#define CUSPARSE_ERR_CHK_NORETURN(stat, tx_function, cusparse_function) if (stat != CUSPARSE_STATUS_SUCCESS) {std::cout<<"Failed in "<<tx_function<<" ... "<<cusparse_function<<" : " <<cusparseGetErrorString(stat)<<std::endl;}
#define CUDA_ERR_CHK_NORETURN(err, tx_function, cuda_function) if (err != cudaSuccess) {std::cout<<"Failed in "<<tx_function<<" ... "<<cuda_function<<" ... cuda error "<<cudaGetErrorString(err) <<std::endl;}

#if THRUST_VERSION >= 100500
#include <thrust/detail/backend/cuda/arch.h>
using namespace thrust::detail::backend::cuda::arch;
#else
#include <thrust/detail/device/cuda/arch.h>
using namespace thrust::detail::device::cuda::arch;
#endif

#define CUSPINTARRAYGPU32 cusp::array1d<int,cusp::device_memory>
#define CUSPINTARRAYCPU32 cusp::array1d<int,cusp::host_memory>

#define DIVIDE_INTO(x,y) ((x + y - 1)/y)

#define MAT_OPERATION cusparseOperation_t 
#define NON_TRANSPOSE CUSPARSE_OPERATION_NON_TRANSPOSE
#define TRANSPOSE CUSPARSE_OPERATION_TRANSPOSE
#define CONJUGATE_TRANSPOSE CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE

template<typename index_type, typename value_type>
  struct cscMatrix {
    index_type num_rows;
    index_type num_cols;
    index_type num_entries;
    CUSPARRAY * values;
    CUSPINTARRAYGPU32 * row_indices;
    CUSPINTARRAYGPU32 * column_offsets;
  };

template<typename index_type, typename value_type>
  struct csrMatrix {
    index_type num_rows;
    index_type num_cols;
    index_type num_entries;
    CUSPARRAY * values;
    CUSPINTARRAYGPU32 * row_offsets;
    CUSPINTARRAYGPU32 * column_indices;
  };


template<typename value_type>
class Equals {
 public:
  __device__ value_type operator() (value_type a, value_type b) {
    return b;
  }
};

template<typename value_type>
class PlusEquals {
 public:
  __device__ value_type operator() (value_type a, value_type b) {
    return a+b;
  }
};


struct ThrustPlusEquals
{
  template <typename Tuple>
  __host__ __device__
  void operator()(Tuple t)
  {
    thrust::get<1>(t) = thrust::get<1>(t) + thrust::get<0>(t);
  }
};


template <typename value_type>
__host__ __device__ static __inline__ value_type _fma_(const value_type a, value_type b, value_type c){

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_SINGLE)
  return cuCfmaf(a, b, c);
#elif defined(PETSC_USE_REAL_DOUBLE)
  return cuCfma(a, b, c);
#endif

#else
  return a*b + c;
#endif
};

template <typename value_type>
__host__ __device__ static __inline__ value_type _zero_(){

#if defined(PETSC_USE_COMPLEX)

#if defined(PETSC_USE_REAL_SINGLE)
  return make_cuFloatComplex(0.f, 0.f);
#elif defined(PETSC_USE_REAL_DOUBLE)
  return make_cuDoubleComplex(0., 0.);
#endif

#else

#if defined(PETSC_USE_REAL_SINGLE)
  return 0.f;
#elif defined(PETSC_USE_REAL_DOUBLE)
  return 0.;
#endif

#endif
};

#endif // TX_COMMON_H
