/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef TX_CSR_MATRIX_H
#define TX_CSR_MATRIX_H

#include <cusp/csr_matrix.h>
#include <tx_common.h>

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_SINGLE)  
#define cusparse_solve cusparseCcsrsv_solve
#define cusparse_analysis cusparseCcsrsv_analysis
#define cusparse_spmv cusparseCcsrmv
#define cusparse_csr2csc cusparseCcsr2csc
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparse_solve cusparseZcsrsv_solve
#define cusparse_analysis cusparseZcsrsv_analysis
#define cusparse_spmv cusparseZcsrmv
#define cusparse_csr2csc cusparseZcsr2csc
#endif
#else
#if defined(PETSC_USE_REAL_SINGLE)  
#define cusparse_solve cusparseScsrsv_solve
#define cusparse_analysis cusparseScsrsv_analysis
#define cusparse_spmv cusparseScsrmv
#define cusparse_csr2csc cusparseScsr2csc
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparse_solve cusparseDcsrsv_solve
#define cusparse_analysis cusparseDcsrsv_analysis
#define cusparse_spmv cusparseDcsrmv
#define cusparse_csr2csc cusparseDcsr2csc
#endif
#endif

#define CUSPMATRIXCSR cusp::csr_matrix<PetscInt,PetscScalar,cusp::device_memory>
#define CUSPARSEMATRIXCSR csrMatrix<PetscInt,PetscScalar>
#define CUSPARSEMATRIXCSC cscMatrix<PetscInt,PetscScalar>
// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_csr_vector_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2012 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------
#ifdef __CUDACC__
template <typename index_type, typename value_type, unsigned int THREADS_PER_VECTOR, unsigned int NNZ_PER_ROW_MAX, unsigned int NODEMAX>
__global__ void spmv_inode_kernel(const index_type numNodes,
				  const index_type * Ap, // A.row_offsets
				  const index_type * Aj, // A.column_indices
				  const value_type * Ax, // A.values
				  const index_type * inodes,
				  const value_type * x,
				  value_type* y)
{
  __shared__ volatile value_type xdata[NNZ_PER_ROW_MAX];
  __shared__ volatile value_type sdata[NODEMAX*THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];
  __shared__ volatile index_type row_offsets[2];
  __shared__ volatile index_type inode[2];
  
  /* thread id in the inode block */
  const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);
    
  /* vector index within the block */
  const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;

  /* index within the shared memory data structure for the reduction */
  const index_type shared_mem_lane = vector_lane * THREADS_PER_VECTOR + thread_lane;

  for(index_type row = blockIdx.x; row < numNodes; row += gridDim.x) {
    /* Get the inode and the row offsets */
    index_type inode0;
    if (threadIdx.x<2) {
      inode[threadIdx.x]=inodes[row+threadIdx.x];
      row_offsets[threadIdx.x] = Ap[inode[0]+threadIdx.x];
    }
    __syncthreads();
    inode0 = inode[0];

    index_type row_diff   = row_offsets[1]-row_offsets[0];

    /* next, load x into shared memory, xdata */
    if (threadIdx.x < row_diff)
      xdata[threadIdx.x] = x[Aj[row_offsets[0] + threadIdx.x]];
    __syncthreads();

    if (vector_lane < inode[1]-inode0) {
      /* initialize local sum */
      value_type sum = _zero_<value_type>();
      index_type row_shift  = row_offsets[0] + vector_lane*row_diff;
      index_type jj = thread_lane;
      
      /* accumulate local sums */
#if CUDA_VERSION == 4000
      sum += ((jj<row_diff) + 0.*(!(jj<row_diff)))*xdata[jj]*Ax[jj + row_shift];
#else
      if(jj < row_diff)
	sum = _fma_<value_type>(Ax[jj+row_shift], xdata[jj], sum);
      //sum = _add_<value_type>(sum, _mult_<value_type>(Ax[jj+row_shift], xdata[jj]));
      //sum += Ax[jj+row_shift] * xdata[jj];
#endif
      for(jj += THREADS_PER_VECTOR; jj < row_diff; jj += THREADS_PER_VECTOR)
	sum = _fma_<value_type>(Ax[jj+row_shift], xdata[jj], sum);
      //sum = _add_<value_type>(sum, _mult_<value_type>(Ax[jj+row_shift], xdata[jj]));
      //sum += Ax[jj + row_shift] * xdata[jj];

      /* store local sum in shared memory */
      sdata[shared_mem_lane] = sum;

      /* reduce local sums to row sum */
      if(THREADS_PER_VECTOR > 16) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane + 16];
      if(THREADS_PER_VECTOR >  8) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  8];
      if(THREADS_PER_VECTOR >  4) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  4];
      if(THREADS_PER_VECTOR >  2) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  2];
      if(THREADS_PER_VECTOR >  1) sdata[shared_mem_lane] = sum = sum + sdata[shared_mem_lane +  1];

      /* first thread of EACH vector writes the result */
      if(thread_lane==0)
      	y[inode0 + vector_lane] = sum;
    }
  }
}

// -----------------------------------------------------------------------------


template <unsigned int THREADS_PER_VECTOR, unsigned int NNZ_PER_ROW_MAX, typename index_type, typename value_type>
    static void spmv_inode_T(const cudaStream_t stream,
			     const CUSPMATRIXCSR& A, 
			     const index_type numNodes, 
			     const index_type nodeMax,
			     const index_type* inodes, 
			     const value_type* x,
			     value_type* y)
{
  //if (UseCache)
  // bind_x(x);
  if (nodeMax==1) {
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 1>, THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);
    spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 1>
      <<<NUM_BLOCKS, THREADS_PER_VECTOR, 0, stream>>>(numNodes,
						      thrust::raw_pointer_cast(&A.row_offsets[0]),
						      thrust::raw_pointer_cast(&A.column_indices[0]),
						      thrust::raw_pointer_cast(&A.values[0]),
						      inodes, x, y);
  }
  else if (nodeMax==2) {
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 2>, 2*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 2>
      <<<NUM_BLOCKS, 2*THREADS_PER_VECTOR, 0, stream>>>(numNodes,
							thrust::raw_pointer_cast(&A.row_offsets[0]),
							thrust::raw_pointer_cast(&A.column_indices[0]),
							thrust::raw_pointer_cast(&A.values[0]),
							inodes, x, y);
  }
  else if (nodeMax==3) {
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 3>, 3*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 3>
      <<<NUM_BLOCKS, 3*THREADS_PER_VECTOR, 0, stream>>>(numNodes,
							thrust::raw_pointer_cast(&A.row_offsets[0]),
							thrust::raw_pointer_cast(&A.column_indices[0]),
							thrust::raw_pointer_cast(&A.values[0]),
							inodes, x, y);
  }
  else if (nodeMax==4) {
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 4>, 4*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);

    spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 4>
      <<<NUM_BLOCKS, 4*THREADS_PER_VECTOR, 0, stream>>>(numNodes,
							thrust::raw_pointer_cast(&A.row_offsets[0]),
							thrust::raw_pointer_cast(&A.column_indices[0]),
							thrust::raw_pointer_cast(&A.values[0]),
							inodes, x, y);
  }
  else if (nodeMax==5) {
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 5>, 5*THREADS_PER_VECTOR, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, numNodes);
    //const size_t NUM_BLOCKS = std::min<size_t>(112, numNodes);
    spmv_inode_kernel<index_type, value_type, THREADS_PER_VECTOR, NNZ_PER_ROW_MAX, 5>
      <<<NUM_BLOCKS, 5*THREADS_PER_VECTOR, 0, stream>>>(numNodes,
							thrust::raw_pointer_cast(&A.row_offsets[0]),
							thrust::raw_pointer_cast(&A.column_indices[0]),
							thrust::raw_pointer_cast(&A.values[0]),
							inodes, x, y);
  }

  //if (UseCache)
  // unbind_x(x);
}


// -----------------------------------------------------------------------------

template <typename index_type, typename value_type>
  void spmv_inode(const cudaStream_t stream,
		  const CUSPMATRIXCSR& A,
		  const index_type numNodes,
		  const index_type nodeMax,
		  const index_type nnzPerRowMax,
		  const index_type* inodes, 
		  const value_type* x,
		  value_type* y)
{
  const index_type nnz_per_row = A.num_entries / A.num_rows;

#ifdef CUSP_USE_TEXTURE_MEMORY 
  #define USECACHE 1
#else
  #define USECACHE 0
#endif

  if(nnz_per_row <=  2) {
    if (nnzPerRowMax <=  2)    { spmv_inode_T<2,   2>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  4)    { spmv_inode_T<2,   4>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  8)    { spmv_inode_T<2,   8>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  16)   { spmv_inode_T<2,  16>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  32)   { spmv_inode_T<2,  32>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  64)   { spmv_inode_T<2,  64>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  128)  { spmv_inode_T<2, 128>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  256)  { spmv_inode_T<2, 256>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  512)  { spmv_inode_T<2, 512>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if !defined(PETSC_USE_COMPLEX)
    if (nnzPerRowMax <=  1024) { spmv_inode_T<2,1024>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { spmv_inode_T<2,2048>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#endif
#endif
  }
  if(nnz_per_row <=  4) {
    if (nnzPerRowMax <=  4)    { spmv_inode_T<4,   4>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  8)    { spmv_inode_T<4,   8>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  16)   { spmv_inode_T<4,  16>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  32)   { spmv_inode_T<4,  32>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  64)   { spmv_inode_T<4,  64>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  128)  { spmv_inode_T<4, 128>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  256)  { spmv_inode_T<4, 256>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  512)  { spmv_inode_T<4, 512>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if !defined(PETSC_USE_COMPLEX)
    if (nnzPerRowMax <=  1024) { spmv_inode_T<4,1024>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { spmv_inode_T<4,2048>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#endif
#endif
  }
  if(nnz_per_row <=  8) {
    if (nnzPerRowMax <=  8)    { spmv_inode_T<8,   8>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  16)   { spmv_inode_T<8,  16>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  32)   { spmv_inode_T<8,  32>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  64)   { spmv_inode_T<8,  64>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  128)  { spmv_inode_T<8, 128>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  256)  { spmv_inode_T<8, 256>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  512)  { spmv_inode_T<8, 512>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if !defined(PETSC_USE_COMPLEX)
    if (nnzPerRowMax <=  1024) { spmv_inode_T<8,1024>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { spmv_inode_T<8,2048>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#endif
#endif
  }
  if(nnz_per_row <= 16) {
    if (nnzPerRowMax <=  16)   { spmv_inode_T<16,  16>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  32)   { spmv_inode_T<16,  32>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  64)   { spmv_inode_T<16,  64>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  128)  { spmv_inode_T<16, 128>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  256)  { spmv_inode_T<16, 256>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  512)  { spmv_inode_T<16, 512>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if !defined(PETSC_USE_COMPLEX)
    if (nnzPerRowMax <=  1024) { spmv_inode_T<16,1024>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { spmv_inode_T<16,2048>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#endif
#endif
  }
  else {
    if (nnzPerRowMax <=  32)   { spmv_inode_T<32,  32>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  64)   { spmv_inode_T<32,  64>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  128)  { spmv_inode_T<32, 128>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  256)  { spmv_inode_T<32, 256>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
    if (nnzPerRowMax <=  512)  { spmv_inode_T<32, 512>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if !defined(PETSC_USE_COMPLEX)
    if (nnzPerRowMax <=  1024) { spmv_inode_T<32,1024>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#if __CUDA_ARCH__ >= 200
    if (nnzPerRowMax <=  2048) { spmv_inode_T<32,2048>(stream, A, numNodes, nodeMax, inodes, x, y); return;}
#endif
#endif
  }
  return;
}


//////////////////////////////////////////////////////////////////////////////
// CSR SpMV kernels based on a vector model (one warp per row)
//////////////////////////////////////////////////////////////////////////////
//
// spmv_csr_vector_device
//   Each row of the CSR matrix is assigned to a warp.  The warp computes
//   y[i] = A[i,:] * x, i.e. the dot product of the i-th row of A with 
//   the x vector, in parallel.  This division of work implies that 
//   the CSR index and data arrays (Aj and Ax) are accessed in a contiguous
//   manner (but generally not aligned).  On GT200 these accesses are
//   coalesced, unlike kernels based on the one-row-per-thread division of 
//   work.  Since an entire 32-thread warp is assigned to each row, many 
//   threads will remain idle when their row contains a small number 
//   of elements.  This code relies on implicit synchronization among 
//   threads in a warp.
//
// spmv_csr_vector_tex_device
//   Same as spmv_csr_vector_tex_device, except that the texture cache is 
//   used for accessing the x vector.
//  
//  Note: THREADS_PER_VECTOR must be one of [2,4,8,16,32]


template <typename index_type, typename value_type, unsigned int VECTORS_PER_BLOCK, unsigned int THREADS_PER_VECTOR, bool UseCache, class OPERATOR>
  __launch_bounds__(VECTORS_PER_BLOCK * THREADS_PER_VECTOR,1)
  __global__ void spmv_csr_vector_kernel(const index_type num_rows,
				       const index_type * Ap, 
				       const index_type * Aj, 
				       const value_type * Ax, 
				       const value_type * x, 
				       value_type * y, OPERATOR OP)
{
    __shared__ volatile value_type sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
    __shared__ volatile index_type ptrs[VECTORS_PER_BLOCK][2];
    
    const index_type THREADS_PER_BLOCK = VECTORS_PER_BLOCK * THREADS_PER_VECTOR;

    const index_type thread_id   = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
    const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
    const index_type vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
    const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
    const index_type num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

    for(index_type row = vector_id; row < num_rows; row += num_vectors)
    {
        // use two threads to fetch Ap[row] and Ap[row+1]
        // this is considerably faster than the straightforward version
        if(thread_lane < 2)
            ptrs[vector_lane][thread_lane] = Ap[row + thread_lane];

        const index_type row_start = ptrs[vector_lane][0];                   //same as: row_start = Ap[row];
        const index_type row_end   = ptrs[vector_lane][1];                   //same as: row_end   = Ap[row+1];

        // initialize local sum
        value_type sum = _zero_<value_type>();
     
        if (THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
        {
            // ensure aligned memory access to Aj and Ax

            index_type jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;

            // accumulate local sums
            if(jj >= row_start && jj < row_end)
	      sum = _fma_<value_type>(Ax[jj], x[Aj[jj]], sum);
	    //sum += _mult_<value_type>(Ax[jj], x[Aj[jj]]);
	    //                sum += Ax[jj] * x[Aj[jj]];

            // accumulate local sums
            for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
	      sum = _fma_<value_type>(Ax[jj], x[Aj[jj]], sum);
	    //sum += _mult_<value_type>(Ax[jj], x[Aj[jj]]);
	    //                sum += Ax[jj] * x[Aj[jj]];
        }
        else
        {
            // accumulate local sums
            for(index_type jj = row_start + thread_lane; jj < row_end; jj += THREADS_PER_VECTOR)
	      sum = _fma_<value_type>(Ax[jj], x[Aj[jj]], sum);
	    //sum += _mult_<value_type>(Ax[jj], x[Aj[jj]]);
	    //                sum += Ax[jj] * x[Aj[jj]];
        }

        // store local sum in shared memory
        sdata[threadIdx.x] = sum;
        
        // reduce local sums to row sum
        if (THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
        if (THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
        if (THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
        if (THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
        if (THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];
       
        // first thread writes the result
        if (thread_lane == 0)
	  y[row] = OP(y[row], sdata[threadIdx.x]);
	//y[row] = sdata[threadIdx.x];
    }
}

template <bool UseCache, unsigned int THREADS_PER_VECTOR, typename index_type, typename value_type, class OPERATOR>
void __spmv_csr_vector(const cudaStream_t stream,
		       const CUSPMATRIXCSR& A, 
		       const value_type* x, 
		       value_type* y, OPERATOR OP)
{
    const size_t THREADS_PER_BLOCK  = 128;
    const size_t VECTORS_PER_BLOCK  = THREADS_PER_BLOCK / THREADS_PER_VECTOR;

    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_csr_vector_kernel<index_type, value_type, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache, OPERATOR>, THREADS_PER_BLOCK, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, DIVIDE_INTO(A.num_rows, VECTORS_PER_BLOCK));
    
    //if (UseCache)
    // bind_x(x);

    spmv_csr_vector_kernel<index_type, value_type, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache> <<<NUM_BLOCKS, THREADS_PER_BLOCK, 0, stream>>> 
        (A.num_rows,
	 thrust::raw_pointer_cast(&A.row_offsets[0]),
	 thrust::raw_pointer_cast(&A.column_indices[0]),
	 thrust::raw_pointer_cast(&A.values[0]),
         x, y, OP);

    //if (UseCache)
    // unbind_x(x);
}

template <typename index_type, typename value_type, class OPERATOR>
  void spmv_csr_vector(const cudaStream_t stream,
		      const CUSPMATRIXCSR& A, 
		      const value_type* x, 
		      value_type* y, 
		      OPERATOR OP)
{
#ifdef CUSP_USE_TEXTURE_MEMORY 
  #define USECACHE 1
#else
  #define USECACHE 0
#endif

  const index_type nnz_per_row = A.num_entries / A.num_rows;
  
  if (nnz_per_row <=  2) { __spmv_csr_vector<USECACHE, 2, index_type, value_type>(stream, A, x, y, OP); return; }
  if (nnz_per_row <=  4) { __spmv_csr_vector<USECACHE, 4, index_type, value_type>(stream, A, x, y, OP); return; }
  if (nnz_per_row <=  8) { __spmv_csr_vector<USECACHE, 8, index_type, value_type>(stream, A, x, y, OP); return; }
  if (nnz_per_row <= 16) { __spmv_csr_vector<USECACHE,16, index_type, value_type>(stream, A, x, y, OP); return; }    
  __spmv_csr_vector<USECACHE,32, index_type, value_type>(stream, A, x, y, OP);
  return;
}

#if defined(PETSC_USE_COMPLEX)
struct conjugate 
{
  __host__ __device__
  PetscScalar operator()(PetscScalar x)
  {
    return cusp::conj(x);
  }
};
#endif

/***********************/
/*   CSR Matrix        */
/***********************/
template <typename index_type, typename value_type>
class CsrMatrix {
public:
  typedef CUSPMATRIXCSR MatrixType;
  typedef CUSPARSEMATRIXCSR MatrixTypeCSR;
  typedef CUSPARSEMATRIXCSC MatrixTypeCSC;

  /* Constructor */
  CsrMatrix() {
    // set points to 0
    inodes=0;

    // initialize other data to 0
    nnzPerRowMax=0;
    nodeMax=0;
    useInode=0;
    nodeCount=0;

    csrA=0;
    cscA=0;
  }

  /* Destructor */
  ~CsrMatrix() {
    if (inodes)
      delete inodes;
    if (csrA) {
      delete csrA->values;
      delete csrA->row_offsets;
      delete csrA->column_indices;
      delete csrA;
      csrA=0;
    }
    if (cscA) {
      delete cscA->values;
      delete cscA->row_indices;
      delete cscA->column_offsets;
      delete cscA;
      cscA=0;
    }
  }

  // Get the matrix type;
  Tx_Mat_type_t getMatrixType() {
    return TX_CSR;
  }

  /* Set the Matrix explicitly on the device*/
  void setMatrix(const cusp::csr_matrix<index_type,value_type,cusp::host_memory>& matCPU) {
    A = matCPU;
  }  

  /* Set the Matrix */
  cusparseStatus_t setCusparseMatrix(cusparseHandle_t cusparseHandle, cusparseMatDescr_t cusparseMatDescription,
				     index_type numRows, index_type numCols, index_type nz,
				     index_type * row_offsets, index_type * column_indices,
				     value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!csrA) {
      // Build csc matrix on demand
      csrA = new MatrixTypeCSR;
      
      csrA->num_rows = numRows;
      csrA->num_cols = numRows;
      csrA->num_entries = nz;
      
      csrA->values = new CUSPARRAY;
      csrA->values->assign(values, values+nz);

      csrA->row_offsets = new CUSPINTARRAYGPU32;
      csrA->row_offsets->assign(row_offsets, row_offsets+numRows+1);
      
      csrA->column_indices = new CUSPINTARRAYGPU32;
      csrA->column_indices->assign(column_indices, column_indices+nz);
    }
    return stat;
  }

  /* Set the Matrix */
  cusparseStatus_t setCusparseMatrixTranspose(cusparseHandle_t cusparseHandle, cusparseMatDescr_t cusparseMatDescription,
				     index_type numRows, index_type numCols, index_type nz,
				     index_type * row_offsets, index_type * column_indices,
				     value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cscA && csrA) {
      // Build csc matrix on demand
      cscA = new MatrixTypeCSC;
      cscA->num_rows = csrA->num_cols;
      cscA->num_cols = csrA->num_rows;
      cscA->num_entries = csrA->num_entries;
      
      cscA->values = new CUSPARRAY;
      cscA->values->resize(csrA->num_entries);
      
      cscA->row_indices = new CUSPINTARRAYGPU32;
      cscA->row_indices->resize(csrA->num_entries);
      
      cscA->column_offsets = new CUSPINTARRAYGPU32;
      cscA->column_offsets->resize(csrA->num_cols+1);
      
      stat=cusparse_csr2csc(cusparseHandle, csrA->num_rows, csrA->num_cols, csrA->num_entries,
			    thrust::raw_pointer_cast(csrA->values->data()),
			    thrust::raw_pointer_cast(csrA->row_offsets->data()),
			    thrust::raw_pointer_cast(csrA->column_indices->data()),
			    thrust::raw_pointer_cast(cscA->values->data()), 
			    thrust::raw_pointer_cast(cscA->row_indices->data()), 
			    thrust::raw_pointer_cast(cscA->column_offsets->data()),
			    CUSPARSE_ACTION_NUMERIC, CUSPARSE_INDEX_BASE_ZERO);
      CUSPARSE_ERR_CHK(stat, "setCusparseMatrixTranspose()", "cusparse_csr2csc()")
    }
    return stat;
  }


  /* Set the Matrix */
  cudaError_t setMatrix(index_type numRows, index_type numCols, index_type nz,
			index_type * row_offsets, index_type * column_indices,
			value_type * values) {

    // easiest way of doing things.
    A.resize(numRows,numCols,nz);
    A.row_offsets.assign(row_offsets, row_offsets+numRows+1);
    A.column_indices.assign(column_indices, column_indices+nz);
    A.values.assign(values, values+nz);
    return cudaGetLastError();
  }

  /* Get the Matrix */
  virtual int getCsrMatrix(CUSPMATRIXCSR** Mat) {
    *Mat = &A;
    return 0;
  }

  /* Set the inode data for when inode.use==true (inodes) */
  void setInodeData(const index_type * inodesCPU, const int inodeCountPlusOne, const int nnzPerRowMaxCPU,
                            const int nodeMaxCPU, const int nodeCountCPU) {

    inodes = new CUSPINTARRAYGPU;
    inodes->assign(inodesCPU, inodesCPU+inodeCountPlusOne);
    nnzPerRowMax = nnzPerRowMaxCPU;
    nodeMax = nodeMaxCPU;
    nodeCount = nodeCountCPU;
    useInode = 1;
  }

  /* cusparse SpMV */
  int cusparseMultiply(const cusparseHandle_t cusparseHandle, const cusparseMatDescr_t cusparseMatDescription,
		       const CUSPARRAY *x, CUSPARRAY *y) {
    /* Use CUSPARSE */
    cusparseStatus_t stat;

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
    cuDoubleComplex alpha = {1.0, 0.0};
    cuDoubleComplex beta = {0.0, 0.0};
#else
    cuFloatComplex alpha = {1.0f, 0.0f};
    cuFloatComplex beta = {0.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
    PetscScalar alpha = 1.0;
    PetscScalar beta = 0.0;
#else
    PetscScalar alpha = 1.0f;
    PetscScalar beta = 0.0f;
#endif
#endif

    if (csrA) {
      stat = cusparse_spmv(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			   csrA->num_rows, csrA->num_cols, csrA->num_entries,
			   &alpha, cusparseMatDescription, 
			   thrust::raw_pointer_cast(csrA->values->data()),
			   thrust::raw_pointer_cast(csrA->row_offsets->data()),
			   thrust::raw_pointer_cast(csrA->column_indices->data()),
			   thrust::raw_pointer_cast(x->data()), &beta,			    
			   thrust::raw_pointer_cast(y->data()));    
      CUSPARSE_ERR_CHK(stat, "cusparseMultiply()", "cusparse_spmv()")
    }
    return (int) stat;
  }

  /* cusparse SpMV^T */
  int cusparseMultiplyTranspose(const cusparseHandle_t cusparseHandle, 
				const cusparseMatDescr_t cusparseMatDescription,
				const CUSPARRAY *x, CUSPARRAY *y) {
    /* Use CUSPARSE */
    cusparseStatus_t stat;

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
    cuDoubleComplex alpha = {1.0, 0.0};
    cuDoubleComplex beta = {0.0, 0.0};
#else
    cuFloatComplex alpha = {1.0f, 0.0f};
    cuFloatComplex beta = {0.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
    PetscScalar alpha = 1.0;
    PetscScalar beta = 0.0;
#else
    PetscScalar alpha = 1.0f;
    PetscScalar beta = 0.0f;
#endif
#endif

    if (cscA) {
      stat = cusparse_spmv(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,			   
			   cscA->num_rows, cscA->num_cols, cscA->num_entries,
			   &alpha, cusparseMatDescription, 
			   thrust::raw_pointer_cast(cscA->values->data()), 
			   thrust::raw_pointer_cast(cscA->column_offsets->data()), 
			   thrust::raw_pointer_cast(cscA->row_indices->data()),
			   thrust::raw_pointer_cast(x->data()), &beta,			    
			   thrust::raw_pointer_cast(y->data()));    	
      CUSPARSE_ERR_CHK(stat, "cusparseMultiplyTranspose()", "cusparse_spmv()")
    }
    return (int) stat;
  }


  /* SpMV with our CUSP-like kernels */
  int multiply(const cudaStream_t stream, const CUSPARRAY *x, CUSPARRAY *y) {

    if (useInode)
      spmv_inode<index_type, value_type>(stream, A, nodeCount, nodeMax, nnzPerRowMax,
					 thrust::raw_pointer_cast(inodes->data()), 
					 thrust::raw_pointer_cast(x->data()), 
					 thrust::raw_pointer_cast(y->data()));
    else
      spmv_csr_vector<index_type, value_type>(stream, A, 
					      thrust::raw_pointer_cast(x->data()), 
					      thrust::raw_pointer_cast(y->data()), 
					      Equals<value_type>());
    return (int) cudaGetLastError();
  }


  /* Solve Analysis */
  cusparseStatus_t solveAnalysis(cusparseHandle_t cusparseHandle,
				 cusparseMatDescr_t cusparseMatDescription,
				 cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
				 MAT_OPERATION transA) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (transA==TRANSPOSE) {
      if (!cscA) {
	// Build csc matrix on demand
	cscA = new MatrixTypeCSC;
	cscA->num_rows = csrA->num_cols;
	cscA->num_cols = csrA->num_rows;
	cscA->num_entries = csrA->num_entries;

	cscA->values = new CUSPARRAY;
	cscA->values->resize(cscA->num_entries);

	cscA->row_indices = new CUSPINTARRAYGPU32;
	cscA->row_indices->resize(cscA->num_entries);

	cscA->column_offsets = new CUSPINTARRAYGPU32;
	cscA->column_offsets->resize(cscA->num_rows+1);

	stat=cusparse_csr2csc(cusparseHandle, csrA->num_rows, csrA->num_cols, csrA->num_entries,
			      thrust::raw_pointer_cast(csrA->values->data()),
			      thrust::raw_pointer_cast(csrA->row_offsets->data()),
			      thrust::raw_pointer_cast(csrA->column_indices->data()),
			      thrust::raw_pointer_cast(cscA->values->data()), 
			      thrust::raw_pointer_cast(cscA->row_indices->data()), 
			      thrust::raw_pointer_cast(cscA->column_offsets->data()),
			      CUSPARSE_ACTION_NUMERIC, CUSPARSE_INDEX_BASE_ZERO);
	stat = cusparse_analysis(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
				 cscA->num_rows, cscA->num_entries, cusparseMatDescription, 
				 thrust::raw_pointer_cast(cscA->values->data()),
				 thrust::raw_pointer_cast(cscA->column_offsets->data()),
				 thrust::raw_pointer_cast(cscA->row_indices->data()),
				 cusparseSolveAnalysisInfo);
	CUSPARSE_ERR_CHK(stat, "solveAnalysis()", "cusparse_analysis()")
      }
    }
    else {
      stat = cusparse_analysis(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			       csrA->num_rows, csrA->num_entries, cusparseMatDescription, 
			       thrust::raw_pointer_cast(csrA->values->data()),
			       thrust::raw_pointer_cast(csrA->row_offsets->data()),
			       thrust::raw_pointer_cast(csrA->column_indices->data()),
			       cusparseSolveAnalysisInfo);
      CUSPARSE_ERR_CHK(stat, "solveAnalysis()", "cusparse_analysis()")
    }
    return stat;
  };
  

  /* solve with matrix reordering */
  cusparseStatus_t solve(cusparseHandle_t cusparseHandle,
			 cusparseMatDescr_t cusparseMatDescription,
			 cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
			 const CUSPARRAY *bGPU, CUSPARRAY *xGPU, MAT_OPERATION transA) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
  cuDoubleComplex alpha = {1.0, 0.0};
#else
  cuFloatComplex alpha = {1.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
  PetscScalar alpha = 1.0;
#else
  PetscScalar alpha = 1.0f;
#endif
#endif

    if (transA==TRANSPOSE) {
      stat = cusparse_solve(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			    cscA->num_rows, &alpha, cusparseMatDescription,
			    thrust::raw_pointer_cast(cscA->values->data()),
			    thrust::raw_pointer_cast(cscA->column_offsets->data()),
			    thrust::raw_pointer_cast(cscA->row_indices->data()),
			    cusparseSolveAnalysisInfo,
			    (const PetscScalar *) thrust::raw_pointer_cast(bGPU->data()),
			    (PetscScalar *) thrust::raw_pointer_cast(xGPU->data()));
      CUSPARSE_ERR_CHK(stat, "solve()", "cusparse_solve()")
    }
    else {
      stat = cusparse_solve(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			    csrA->num_rows, &alpha, cusparseMatDescription,
			    thrust::raw_pointer_cast(csrA->values->data()),
			    thrust::raw_pointer_cast(csrA->row_offsets->data()),
			    thrust::raw_pointer_cast(csrA->column_indices->data()),
			    cusparseSolveAnalysisInfo,
			    (const PetscScalar *) thrust::raw_pointer_cast(bGPU->data()),
			    (PetscScalar *) thrust::raw_pointer_cast(xGPU->data()));
      CUSPARSE_ERR_CHK(stat, "solve()", "cusparse_solve()")
    }
    return stat;
  };


 private:
  MatrixType A;/* CUSP container for the matrix */
  MatrixTypeCSC * cscA; /* CUSPARSE container for the transpose of the matrix */
  MatrixTypeCSR * csrA; /* CUSPARSEcontainer for the matrix */

  /* Data below is for inodes */
  CUSPINTARRAYGPU*    inodes; /*pointer to an array containing the inode data structure */
  index_type          useInode;
  index_type          nodeCount; /* largest inode size */
  index_type          nnzPerRowMax; /* maximum number of nonzeros in a row ... for shared memory vector size */
  index_type          nodeMax; /* largest inode size */

};
#endif
#endif // TX_CSR_MATRIX_H
