/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef TX_ELL_MATRIX_H
#define TX_ELL_MATRIX_H

#include <cusp/csr_matrix.h>
#include <cusp/ell_matrix.h>
#include <tx_common.h>

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_SINGLE)  
#define cusparse_ell_solve cusparseChybsv_solve
#define cusparse_ell_analysis cusparseChybsv_analysis
#define cusparse_ell_spmv cusparseChybmv
#define cusparse_ell_convert cusparseCcsr2hyb
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparse_ell_solve cusparseZhybsv_solve
#define cusparse_ell_analysis cusparseZhybsv_analysis
#define cusparse_ell_spmv cusparseZhybmv
#define cusparse_ell_convert cusparseZcsr2hyb
#endif
#else
#if defined(PETSC_USE_REAL_SINGLE)  
#define cusparse_ell_solve cusparseShybsv_solve
#define cusparse_ell_analysis cusparseShybsv_analysis
#define cusparse_ell_spmv cusparseShybmv
#define cusparse_ell_convert cusparseScsr2hyb
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparse_ell_solve cusparseDhybsv_solve
#define cusparse_ell_analysis cusparseDhybsv_analysis
#define cusparse_ell_spmv cusparseDhybmv
#define cusparse_ell_convert cusparseDcsr2hyb
#endif
#endif

#define CUSPMATRIXELL cusp::ell_matrix<PetscInt,PetscScalar,cusp::device_memory>

// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_ell_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2012 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------

#define CUSPMATRIXELL cusp::ell_matrix<PetscInt,PetscScalar,cusp::device_memory>

template <typename index_type, typename value_type, unsigned int BLOCK_SIZE, class OPERATOR>
__launch_bounds__(BLOCK_SIZE,1)
__global__ void spmv_ell_kernel(const index_type num_rows, 
				const index_type num_cols, 
				const index_type num_cols_per_row,
				const index_type pitch,
				const index_type * Aj,
				const value_type * Ax, 
				const value_type * x, 
				value_type * y,
				OPERATOR OP)
{
    const index_type invalid_index = cusp::ell_matrix<index_type, value_type, cusp::device_memory>::invalid_index;

    const index_type thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const index_type grid_size = gridDim.x * blockDim.x;

    for(index_type row = thread_id; row < num_rows; row += grid_size)
    {
      value_type sum = _zero_<value_type>();
      
        index_type offset = row;

        for(index_type n = 0; n < num_cols_per_row; n++)
        {
            const index_type col = Aj[offset];

            if (col != invalid_index)
            {
                const value_type A_ij = Ax[offset];
		sum = _fma_<value_type>(A_ij, x[col], sum);
                //sum += A_ij * x[col];
                //sum += A_ij * fetch_x<UseCache>(col, x);
            }

            offset += pitch;
        }
	y[row] = OP(y[row],sum);
    }
}


template <typename index_type, typename value_type, class OPERATOR>
void spmv_ell(const cudaStream_t stream,
	      const CUSPMATRIXELL& A,
	      const value_type* x, 
	      value_type* y,
	      OPERATOR OP)
{
  const size_t BLOCK_SIZE = 256;
  const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_ell_kernel<index_type,value_type,BLOCK_SIZE, OPERATOR>, BLOCK_SIZE, (size_t) 0);
  const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, DIVIDE_INTO(A.num_rows, BLOCK_SIZE));
  
  const index_type pitch               = A.column_indices.pitch;
  const index_type num_entries_per_row = A.column_indices.num_cols;
  
  // TODO generalize this
  assert(A.column_indices.pitch == A.values.pitch);
  
  //if (UseCache)
  //  bind_x(x);
  
  spmv_ell_kernel<index_type,value_type,BLOCK_SIZE> <<<NUM_BLOCKS, BLOCK_SIZE, 0, stream>>>
    (A.num_rows, A.num_cols,
     num_entries_per_row, pitch,
     thrust::raw_pointer_cast(&A.column_indices.values[0]), 
     thrust::raw_pointer_cast(&A.values.values[0]),
     x, y, OP);
  
  //if (UseCache)
  //  unbind_x(x);
}


/***********************/
/*      Ell Matrix     */
/***********************/
template <typename index_type, typename value_type>
class EllMatrix {
public:
  typedef CUSPMATRIXELL MatrixType;

  /* Constructor */
  EllMatrix() {
    cusparseAIsAllocated=false;
    cusparseATransposeIsAllocated=false;
  }

  /* Destructor */
  ~EllMatrix() {
    if(cusparseAIsAllocated) {
      cusparseStatus_t stat = cusparseDestroyHybMat(cusparseA);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~EllMatrix()", "cusparseDestroyHybMat()")
    }      
    if(cusparseATransposeIsAllocated) {
      cusparseStatus_t stat = cusparseDestroyHybMat(cusparseATranspose);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~EllMatrix()", "cusparseDestroyHybMat()")
    }      
  }

  // Get the matrix type;
  Tx_Mat_type_t getMatrixType() {
    return TX_ELL;
  }

  /* Set the Matrix explicitly on the device*/
  void setMatrix(const cusp::csr_matrix<index_type,value_type,cusp::host_memory>& matCPU) {
    A = matCPU;
  }  

  /* Set the Matrix */
  cusparseStatus_t setCusparseMatrix(cusparseHandle_t cusparseHandle, cusparseMatDescr_t cusparseMatDescription,
			 index_type numRows, index_type numCols, index_type nz,
			 index_type * row_offsets, index_type * column_indices,
			 value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cusparseAIsAllocated) {
      stat = cusparseCreateHybMat(&cusparseA);
      cusparseAIsAllocated=true;
      CUSPARSE_ERR_CHK(stat, "EllMatrix::setCusparseMatrix", "cusparseCreateHybMat()");
      // create a temporary 
      CUSPMATRIXCSR B; /* CSR version of the matrix */
      B.resize(numRows,numCols,nz);    
      B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
      B.column_indices.assign(column_indices, column_indices+nz);
      B.values.assign(values, values+nz);
      
#if defined (PETSC_USE_64BIT_INDICES)
      CUSPINTARRAYGPU32 row_offsets_int = B.row_offsets;
      CUSPINTARRAYGPU32 column_indices_int = B.column_indices;
#endif
      stat = cusparse_ell_convert(cusparseHandle, numRows, numCols, 
				  cusparseMatDescription,
				  thrust::raw_pointer_cast(&B.values[0]),
#if defined (PETSC_USE_64BIT_INDICES)
				  thrust::raw_pointer_cast(&row_offsets_int[0]), 
				  thrust::raw_pointer_cast(&column_indices_int[0]),
#else
				  thrust::raw_pointer_cast(&B.row_offsets[0]),
				  thrust::raw_pointer_cast(&B.column_indices[0]),
#endif
				  cusparseA, 0, CUSPARSE_HYB_PARTITION_MAX);
      CUSPARSE_ERR_CHK(stat, "EllMatrix::setCusparseMatrix", "cusparse_ell_convert()");
    }
    return stat;
  }

  /* Set the Matrix */
  cusparseStatus_t setCusparseMatrixTranspose(cusparseHandle_t cusparseHandle, cusparseMatDescr_t cusparseMatDescription,
					      index_type numRows, index_type numCols, index_type nz,
					      index_type * row_offsets, index_type * column_indices,
					      value_type * values) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    if (!cusparseATransposeIsAllocated) {
      stat = cusparseCreateHybMat(&cusparseATranspose);
      cusparseATransposeIsAllocated=true;
      CUSPARSE_ERR_CHK(stat, "EllMatrix::setCusparseMatrixTranspose", "cusparseCreateHybMat()");
      // create a temporary 
      CUSPMATRIXCSR B; /* CSR version of the matrix */
      B.resize(numRows,numCols,nz);    
      B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
      B.column_indices.assign(column_indices, column_indices+nz);
      B.values.assign(values, values+nz);
      
#if defined (PETSC_USE_64BIT_INDICES)
      CUSPINTARRAYGPU32 row_offsets_int = B.row_offsets;
      CUSPINTARRAYGPU32 column_indices_int = B.column_indices;
#endif
      
      // Build csc matrix on demand    
      CUSPARRAY cscValues(nz);
      CUSPINTARRAYGPU32 cscRowIndices(nz);
      CUSPINTARRAYGPU32 cscColumnOffsets(numCols+1);    
      stat=cusparse_csr2csc(cusparseHandle, numRows, numCols, nz,
			    thrust::raw_pointer_cast(B.values.data()),
#if defined (PETSC_USE_64BIT_INDICES)
			    thrust::raw_pointer_cast(&row_offsets_int[0]), 
			    thrust::raw_pointer_cast(&column_indices_int[0]),
#else
			    thrust::raw_pointer_cast(B.row_offsets.data()),
			    thrust::raw_pointer_cast(B.column_indices.data()),
#endif
			    thrust::raw_pointer_cast(cscValues.data()), 
			    thrust::raw_pointer_cast(cscRowIndices.data()), 
			    thrust::raw_pointer_cast(cscColumnOffsets.data()),
			    CUSPARSE_ACTION_NUMERIC, CUSPARSE_INDEX_BASE_ZERO);
      
      
      stat = cusparse_ell_convert(cusparseHandle, numRows, numCols, 
				  cusparseMatDescription,
				  thrust::raw_pointer_cast(cscValues.data()), 
				  thrust::raw_pointer_cast(cscColumnOffsets.data()),
				  thrust::raw_pointer_cast(cscRowIndices.data()), 
				  cusparseATranspose, 0, CUSPARSE_HYB_PARTITION_MAX);
      CUSPARSE_ERR_CHK(stat, "EllMatrix::setCusparseMatrixTranpose", "cusparse_ell_convert()");
    }
    return stat;  
  }

  /* Set the Matrix */
  cudaError_t setMatrix(index_type numRows, index_type numCols, index_type nz,
			index_type * row_offsets, index_type * column_indices,
			value_type * values) {

    // create a temporary
    printf("setMatrix!\n");
    CUSPMATRIXCSR B; /* CSR version of the matrix */
    B.resize(numRows,numCols,nz);    
    B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
    B.column_indices.assign(column_indices, column_indices+nz);
    B.values.assign(values, values+nz);
    // Let CUSP do the conversion
    A = B;
    return cudaGetLastError();
  }

  /* Get the Matrix */
  virtual int getCsrMatrix(CUSPMATRIXCSR** A) {
    CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "GPU_Matrix::getCsrMatrix()", "not supported for Ell matrix type");
  }

  /* Set the Indices For cprow==true (compressed row) */
  void setCPRowIndices(const index_type * indices, const index_type m) {
    // does nothing as of now
  }

  /* Set the inode data for when inode.use==true (inodes) */
  void setInodeData(const index_type * inodesCPU, const int inodeCountPlusOne, const int nnzPerRowMaxCPU,
                            const int nodeMaxCPU, const int nodeCountCPU) {
    // does nothing as of now
  }

  /* cusparse SpMV */
  int cusparseMultiply(const cusparseHandle_t cusparseHandle, const cusparseMatDescr_t cusparseMatDescription,
		       const CUSPARRAY *x, CUSPARRAY *y) {
    
    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
    cuDoubleComplex alpha = {1.0, 0.0};
    cuDoubleComplex beta = {0.0, 0.0};
#else
    cuFloatComplex alpha = {1.0f, 0.0f};
    cuFloatComplex beta = {0.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
    PetscScalar alpha = 1.0;
    PetscScalar beta = 0.0;
#else
    PetscScalar alpha = 1.0f;
    PetscScalar beta = 0.0f;
#endif
#endif
  
    if (cusparseA)
      stat = cusparse_ell_spmv(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			       &alpha, cusparseMatDescription, cusparseA,
			       thrust::raw_pointer_cast(x->data()), &beta,			    
			       thrust::raw_pointer_cast(y->data()));    
    CUSPARSE_ERR_CHK(stat, "EllMatrix::cusparseMultiply", "cusparse_ell_spmv()");
    return (int) stat;
  }

  /* cusparse SpMV^T */
  int cusparseMultiplyTranspose(const cusparseHandle_t cusparseHandle, const cusparseMatDescr_t cusparseMatDescription,
				const CUSPARRAY *x, CUSPARRAY *y) {

    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
    cuDoubleComplex alpha = {1.0, 0.0};
    cuDoubleComplex beta = {0.0, 0.0};
#else
    cuFloatComplex alpha = {1.0f, 0.0f};
    cuFloatComplex beta = {0.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
    PetscScalar alpha = 1.0;
    PetscScalar beta = 0.0;
#else
    PetscScalar alpha = 1.0f;
    PetscScalar beta = 0.0f;
#endif
#endif
    
    if (cusparseATranspose)
      stat = cusparse_ell_spmv(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
			       &alpha, cusparseMatDescription, cusparseATranspose,
			       thrust::raw_pointer_cast(x->data()), &beta,			    
			       thrust::raw_pointer_cast(y->data()));    
    CUSPARSE_ERR_CHK(stat, "EllMatrix::cusparseMultiply", "cusparse_ell_spmv()");
    return (int) stat;
  }


  /* SpMV with our CUSP-like kernels */
  int multiply(const cudaStream_t stream, const CUSPARRAY *x, CUSPARRAY *y) {
    spmv_ell<index_type, value_type>(stream, A, 
				     thrust::raw_pointer_cast(x->data()), 
				     thrust::raw_pointer_cast(y->data()), 
				     Equals<value_type>());
    CUDA_ERR_CHK(cudaGetLastError(), "EllMatrix::multiply", "spmv_ell()");
    return 0;
  }


  /* Solve Analysis */
  cusparseStatus_t solveAnalysis(cusparseHandle_t cusparseHandle,
				 cusparseMatDescr_t cusparseMatDescription,
				 cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
				 MAT_OPERATION transA) {
    
    if (transA!=CUSPARSE_OPERATION_NON_TRANSPOSE) 
      CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "EllMatrix::solveAnalysis", "cusparse_ell_analysis() ... matrix type not supported for (conjugate) transpose operation.");

    cusparseStatus_t stat = cusparse_ell_analysis(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
						  cusparseMatDescription, cusparseA,
						  cusparseSolveAnalysisInfo);
    CUSPARSE_ERR_CHK(stat, "EllMatrix::solveAnalysis", "cusparse_ell_analysis()");
    return stat;
  };
  

  /* solve */
  cusparseStatus_t solve(cusparseHandle_t cusparseHandle,
			 cusparseMatDescr_t cusparseMatDescription,
			 cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo,
			 const CUSPARRAY *bGPU, CUSPARRAY *xGPU, MAT_OPERATION transA) {

    if (transA!=CUSPARSE_OPERATION_NON_TRANSPOSE) 
      CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "EllMatrix::solve", "cusparse_ell_solve() ... matrix type not supported for (conjugate) transpose operation.");

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_DOUBLE)
  cuDoubleComplex alpha = {1.0, 0.0};
#else
  cuFloatComplex alpha = {1.0f, 0.0f};
#endif
#else
#if defined(PETSC_USE_REAL_DOUBLE)
  PetscScalar alpha = 1.0;
#else
  PetscScalar alpha = 1.0f;
#endif
#endif
    cusparseStatus_t stat = cusparse_ell_solve(cusparseHandle, CUSPARSE_OPERATION_NON_TRANSPOSE,
					       &alpha, cusparseMatDescription, cusparseA,
					       cusparseSolveAnalysisInfo,
					       (const PetscScalar *) thrust::raw_pointer_cast(bGPU->data()),
					       (PetscScalar *) thrust::raw_pointer_cast(xGPU->data()));
    CUSPARSE_ERR_CHK(stat, "EllMatrix::solve", "cusparse_ell_solve()");
    return stat;
  }

private:
  MatrixType A;/* container for the matrix */
  cusparseHybMat_t cusparseA; /* cusparse container for the matrix */
  cusparseHybMat_t cusparseATranspose; /* cusparse container for the matrix transpose */
  bool cusparseAIsAllocated;
  bool cusparseATransposeIsAllocated;
};

#endif // TX_ELL_MATRIX_H
