#if !defined(__TX_SPARSE_INTERFACE_H)
#define __TX_SPARSE_INTERFACE_H

// Matrix type
typedef enum{
  TX_COO=0,
  TX_CSR=1,
  TX_DIA=2,
  TX_ELL=3,
  TX_HYB=4
} Tx_Mat_type_t;

#include "tx_csr_matrix.h"
//#include "tx_coo_matrix.h"
#include "tx_dia_matrix.h"
#include "tx_ell_matrix.h"
#include "tx_hyb_matrix.h"
#include <stdio.h>

#ifndef GPU_TIMER
#define GPU_TIMER
#endif 
#undef GPU_TIMER

//////////////////////////////////////////
// Interface class
//////////////////////////////////////////
class GPU_Matrix_Ifc {
public:

  /* Constructor */
  GPU_Matrix_Ifc() {}

  /* Destructor */
  virtual ~GPU_Matrix_Ifc() {
  }
  
  /* Set the Matrix */
  virtual void setMatrix(const cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory>& matCPU) = 0;

  /* Set the Matrix */
  virtual int setMatrix(PetscInt numRows, PetscInt numCols, PetscInt nz,
			PetscInt * row_offsets, PetscInt * column_indices,
			PetscScalar * values, MAT_OPERATION transA=NON_TRANSPOSE) = 0;
  
  /* Get the Matrix */
  virtual int getCsrMatrix(CUSPMATRIXCSR** A) = 0;
  
  /* Set the Indices For cprow==true (compressed row), m = number of rows */
  virtual void setCPRowIndices(const PetscInt * indices, const int m) = 0;

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodes, const int inodeCountPlusOne, const int nnzPerRowMax,
                            const int nodeMax, const int nodeCount) = 0;

  /* Get the stream for this matrix */
  virtual cudaStream_t getStream() = 0;

  /* build streams and events */
  virtual cudaError_t buildStreamsAndEvents(const PetscInt MPICommSize, cudaStream_t * theBodyStream) = 0;

  /* initialize cusparse for SpMV */
  virtual cusparseStatus_t initializeCusparseMat(cusparseHandle_t MAT_cusparseHandle, 
					      cusparseMatrixType_t MATRIX_TYPE,
					      cusparseFillMode_t UPLO,
					      MAT_OPERATION transA=NON_TRANSPOSE) = 0;

  /* needed test for synchronization */
  virtual bool hasNonZeroStream() = 0;

  /* Multiply */
  virtual int multiply(const CUSPARRAY * x, CUSPARRAY *y, MAT_OPERATION transA=NON_TRANSPOSE) = 0;

  /* Multiply Add */
  virtual int multiplyAdd(const CUSPARRAY * x, CUSPARRAY *y, MAT_OPERATION transA=NON_TRANSPOSE) = 0;

  /* set the indices for matrix reorderings */
  virtual int setOrdIndices(const PetscInt * indicesCPU, PetscInt n) = 0;

  /* reorder bGPU and store the result in xGPU */
  virtual int reorderIn(CUSPARRAY *xGPU, const CUSPARRAY *bGPU) = 0;

  /* reorder xGPU and store the result, ultimately, in xGPU */
  virtual int reorderOut(CUSPARRAY *xGPU) = 0;

  /* Solve Analysis */
  virtual cusparseStatus_t solveAnalysis(MAT_OPERATION transA=NON_TRANSPOSE) = 0;

  /* solve with matrix reordering */
  virtual cusparseStatus_t solve(const CUSPARRAY *bGPU, CUSPARRAY *xGPU, MAT_OPERATION transA=NON_TRANSPOSE) = 0;

  /* get a pointer to the temporary array */
  //virtual int getTempArray(CUSPARRAY** tempGPU) = 0;
};

////////////////////////////////////////////////////
// Derived class templated over matrix type
////////////////////////////////////////////////////
template<class MTYPE>
class GPU_Matrix : public GPU_Matrix_Ifc {
public:
  
  /* Constructor */
  GPU_Matrix() {
    matClass = new MTYPE;
    indices=0;
    
    /* stream stuff */
    stream= 0;
    streamIsACopy=false;

    /* cusparse stuff */
    cusparseHandle=0;
    cusparseMatDescription=0;
    cusparseSolveAnalysisInfo=0;

    cusparseMatDescriptionTranspose=0;
    cusparseSolveAnalysisInfoTranspose=0;

    /* work vector */
    tempArray = new CUSPARRAY;

#ifdef GPU_TIMER
    startGPU=0;
    stopGPU=0;
    dt = 0.f;
    N = 0;
#endif
  }

  /* Destructor */
  virtual ~GPU_Matrix() {
    
    /* CUSPARSE Cleanup */
    cusparseStatus_t stat;    
    if (cusparseMatDescription) {
      stat = cusparseDestroyMatDescr(cusparseMatDescription);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~GPU_Matrix()", "cusparseDestroyMatDescr()")
    }
    if (cusparseSolveAnalysisInfo) {
      stat = cusparseDestroySolveAnalysisInfo(cusparseSolveAnalysisInfo);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~GPU_Matrix()", "cusparseDestroySolveAnalysisInfo()")
    }

    if (cusparseMatDescriptionTranspose) {
      stat = cusparseDestroyMatDescr(cusparseMatDescriptionTranspose);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~GPU_Matrix()", "cusparseDestroyMatDescr()")
    }
    if (cusparseSolveAnalysisInfoTranspose) {
      stat = cusparseDestroySolveAnalysisInfo(cusparseSolveAnalysisInfoTranspose);
      CUSPARSE_ERR_CHK_NORETURN(stat, "~GPU_Matrix()", "cusparseDestroySolveAnalysisInfo()")
    }

    /* Stream Cleanup */
    if (stream && !streamIsACopy) {
      cudaError_t err = cudaStreamDestroy(stream);
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Matrix()", "cudaStreamDestroy()")
      stream=0;
    }
#ifdef GPU_TIMER
    printf("MatMult(Add) Mean dt = %1.5f\t N=%d\n",dt/N, N);    

    // Kill the events
    if (startGPU) {
      cudaError_t err = cudaEventDestroy(startGPU);
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Matrix()", "cudaEventDestroy()")
      startGPU=0;
    }
    if (stopGPU) {
      cudaError_t err = cudaEventDestroy(stopGPU);
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Matrix()", "cudaEventDestroy()")
      stopGPU=0;
    }
#endif
    if (indices)
      delete indices;
    if (tempArray)
      delete tempArray;
    delete matClass;
  }

  /* Set the Handle to the cusparse matrix */
  virtual cusparseStatus_t initializeCusparseMat(cusparseHandle_t MAT_cusparseHandle, 
					      cusparseMatrixType_t MATRIX_TYPE,
					      cusparseFillMode_t UPLO,
					      MAT_OPERATION transA=NON_TRANSPOSE) {

    // These First couple lines are need for both general and triangular matrices
    cusparseStatus_t stat=CUSPARSE_STATUS_SUCCESS;
    cusparseHandle = MAT_cusparseHandle;

    stat = cusparseSetStream(cusparseHandle, stream);
    CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetStream()");

    if (MATRIX_TYPE==CUSPARSE_MATRIX_TYPE_TRIANGULAR) {

      if (transA==NON_TRANSPOSE) {
	stat = cusparseCreateMatDescr(&cusparseMatDescription);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseCreateMatDescr()");

	stat = cusparseSetMatIndexBase(cusparseMatDescription, CUSPARSE_INDEX_BASE_ZERO);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatIndexBase()");
    	
	stat = cusparseSetMatType(cusparseMatDescription, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatType()");

	stat = cusparseSetMatFillMode(cusparseMatDescription, UPLO);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatFillMode()");

	// csr can have unit and non unit diagonals
	if (matClass->getMatrixType()==TX_CSR) {
	  if (UPLO==CUSPARSE_FILL_MODE_UPPER) {
	    stat = cusparseSetMatDiagType(cusparseMatDescription, CUSPARSE_DIAG_TYPE_NON_UNIT);
	  }
	  else {
	    stat = cusparseSetMatDiagType(cusparseMatDescription, CUSPARSE_DIAG_TYPE_UNIT);
	  }
	}
	else {
	  // All other types require non unit diagonal for triangular systems.
	  CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "GPU_Matrix::initializeCusparseMat()", "Only CSR matrices are supported for triangular solve.");
	}      

	stat = cusparseCreateSolveAnalysisInfo(&cusparseSolveAnalysisInfo);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseCreateSolveAnalysisInfo()");
      }
      else if (transA==TRANSPOSE) {
	stat = cusparseCreateMatDescr(&cusparseMatDescriptionTranspose);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseCreateMatDescr()");

	stat = cusparseSetMatIndexBase(cusparseMatDescriptionTranspose, CUSPARSE_INDEX_BASE_ZERO);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatIndexBase()");
	
	stat = cusparseSetMatType(cusparseMatDescriptionTranspose, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatType()");

	stat = cusparseSetMatFillMode(cusparseMatDescriptionTranspose, UPLO);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatFillMode()");

	// csr can have unit and non unit diagonals
	if (matClass->getMatrixType()==TX_CSR) {
	  // csr can have unit and non unit diagonals
	  if (UPLO==CUSPARSE_FILL_MODE_UPPER) 
	    stat = cusparseSetMatDiagType(cusparseMatDescriptionTranspose, CUSPARSE_DIAG_TYPE_UNIT);
	  else
	    stat = cusparseSetMatDiagType(cusparseMatDescriptionTranspose, CUSPARSE_DIAG_TYPE_NON_UNIT);
	}
	else {
	  // All other types require non unit diagonal for triangular systems.
	  CUSPARSE_ERR_CHK(CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED, "GPU_Matrix::initializeCusparseMat()", "Only CSR matrices are supported for triangular solve.");
	}      
      }
    }
    else if (MATRIX_TYPE==CUSPARSE_MATRIX_TYPE_GENERAL) {
      if (transA==NON_TRANSPOSE) {
	stat = cusparseCreateMatDescr(&cusparseMatDescription);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseCreateMatDescr()");
    
	stat = cusparseSetMatIndexBase(cusparseMatDescription, CUSPARSE_INDEX_BASE_ZERO);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatIndexBase()");
	
	stat = cusparseSetMatType(cusparseMatDescription, CUSPARSE_MATRIX_TYPE_GENERAL);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatType()");

      }
      else if (transA==TRANSPOSE) {
	stat = cusparseCreateMatDescr(&cusparseMatDescriptionTranspose);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseCreateMatDescr()");
	
	stat = cusparseSetMatIndexBase(cusparseMatDescriptionTranspose, CUSPARSE_INDEX_BASE_ZERO);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatIndexBase()");

	stat = cusparseSetMatType(cusparseMatDescriptionTranspose, CUSPARSE_MATRIX_TYPE_GENERAL);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::initializeCusparseMat()", "cusparseSetMatType()");
      }
    }
    return stat;
  }


  /* Set the Matrix */
  virtual void setMatrix(const cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory>& matCPU) {
    matClass->setMatrix(matCPU);
  }

  /* Set the Matrix */
  virtual int setMatrix(PetscInt numRows, PetscInt numCols, PetscInt nz,
			 PetscInt * row_offsets, PetscInt * column_indices,
			PetscScalar * values, MAT_OPERATION transA=NON_TRANSPOSE) {

    int err=0;
    if (cusparseHandle) {
      if (transA==NON_TRANSPOSE) {
	// figure out in the class that we templated over
	err = matClass->setCusparseMatrix(cusparseHandle, cusparseMatDescription, 
					  numRows, numCols, nz, row_offsets, column_indices, values);
      }
      else {
	// figure out in the class that we templated over
	err = matClass->setCusparseMatrixTranspose(cusparseHandle, cusparseMatDescriptionTranspose,
						   numRows, numCols, nz, row_offsets, column_indices, values);
      }
    }
    else {
      // figure out in the class that we templated over
      err = matClass->setMatrix(numRows, numCols, nz, row_offsets, column_indices, values);
    }

    /* tempArray */
    tempArray->resize(numRows);
    return err;
  }

  /* Get the Matrix */
  virtual int getCsrMatrix(CUSPMATRIXCSR** A) {
    return matClass->getCsrMatrix(A);
  }

  /* Set the Indices For cprow==true (compressed row), m = number of rows */
  virtual void setCPRowIndices(const PetscInt * cpRowIndices, const int m) {
    indices = new CUSPINTARRAYGPU;
    indices->assign(cpRowIndices,cpRowIndices+m);	
  }

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodes, const int inodeCountPlusOne, const int nnzPerRowMax,
                            const int nodeMax, const int nodeCount) {
    matClass->setInodeData(inodes, inodeCountPlusOne, nnzPerRowMax, nodeMax, nodeCount);
  }

  /* Get the stream for this matrix */
  virtual cudaStream_t getStream() {
    return stream;
  }

  /* build streams and events */
  virtual cudaError_t buildStreamsAndEvents(const PetscInt MPICommSize, cudaStream_t * theStream) {
    if (MPICommSize>1 && !(*theStream)) {
      cudaError_t err = cudaStreamCreate(&stream);
      CUDA_ERR_CHK(err, "GPU_Matrix::buildStreamsAndEvents()", "cudaStreamCreate()")
      *theStream = stream;
    }
    else {
      // set to stream 0 if only 1 MPI rank or the input stream is 0.
      stream = *theStream;
      streamIsACopy=true;
      *theStream=0;
    }
#ifdef GPU_TIMER
    cudaError_t err = cudaEventCreate(&startGPU);
    CUDA_ERR_CHK(err, "GPU_Matrix::buildStreamsAndEvents()", "cudaEventCreate()")
    err = cudaEventCreate(&stopGPU);
    CUDA_ERR_CHK(err, "GPU_Matrix::buildStreamsAndEvents()", "cudaEventCreate()")
    dt=0.f;
    N=0;
#endif
    // success
    return cudaSuccess;
  };

  /* needed test for synchronization */
  virtual bool hasNonZeroStream() {
    if (stream)
      return true;
    else 
      return false;
  }

  /* Multiply */
  virtual int multiply(const CUSPARRAY * x, CUSPARRAY *y, MAT_OPERATION transA=NON_TRANSPOSE) {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    int err;
    if (cusparseHandle) {
      if (transA==NON_TRANSPOSE)
	err = matClass->cusparseMultiply(cusparseHandle, cusparseMatDescription, x, y);
      else
	err = matClass->cusparseMultiplyTranspose(cusparseHandle, cusparseMatDescriptionTranspose, x, y);
    }
    else {
      if (transA!=NON_TRANSPOSE) {
	cudaError_t error = cudaErrorNotYetImplemented;
	CUDA_ERR_CHK(error, "GPU_Matrix::multiply()", "multiply by transpose not supported with CUSP matrices");
      }
      else
	err = matClass->multiply(stream, x, y);
    }    
#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float deltaT=0.f;
    cudaEventElapsedTime(&deltaT, startGPU, stopGPU);	    
    dt+=deltaT;
    N++;
#endif
    return err;
  };

  /* Multiply Add */
  virtual int multiplyAdd(const CUSPARRAY * x, CUSPARRAY *y, MAT_OPERATION transA=NON_TRANSPOSE) {
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    int err;

    /* return if size is zero */
    if (x->size() == 0) return 0;

    if (cusparseHandle) {
      if (transA==NON_TRANSPOSE)
	err = matClass->cusparseMultiply(cusparseHandle, cusparseMatDescription, x, tempArray);
      else
	err = matClass->cusparseMultiplyTranspose(cusparseHandle, cusparseMatDescriptionTranspose, x, tempArray);
    }
    else {
      if (transA!=NON_TRANSPOSE) {
	cudaError_t error = cudaErrorNotYetImplemented;
	CUDA_ERR_CHK(error, "GPU_Matrix::multiplyAdd()", "multiplyAdd by transpose not supported with CUSP matrices");
      }
      else
	err = matClass->multiply(stream, x, tempArray);
    }
    // results are stored in a temporary array above. These are then combined in the correct places.
    thrust::for_each(thrust::make_zip_iterator(thrust::make_tuple(tempArray->begin(), thrust::make_permutation_iterator(y->begin(), indices->begin()))),
		     thrust::make_zip_iterator(thrust::make_tuple(tempArray->begin(), thrust::make_permutation_iterator(y->begin(), indices->begin()))) + tempArray->size(), 
		     ThrustPlusEquals());


#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float deltaT=0.f;
    cudaEventElapsedTime(&deltaT, startGPU, stopGPU);	    
    dt+=deltaT;
    N++;
#endif
   return err;   
  };

  /* set the indices for matrix reorderings */
  virtual int setOrdIndices(const PetscInt * indicesCPU, PetscInt n) {
    indices = new CUSPINTARRAYGPU;
    indices->assign(indicesCPU, indicesCPU+n);
    return 0;
  }

  /* reorder bGPU and store the result in xGPU */
  virtual int reorderIn(CUSPARRAY *xGPU, const CUSPARRAY *bGPU) {
    // Copy the right hand side vector, bGPU, into xGPU with the row permutation
    thrust::copy(thrust::make_permutation_iterator(bGPU->begin(), indices->begin()),
		 thrust::make_permutation_iterator(bGPU->end(), indices->end()),
		 xGPU->begin());
    return 0;
  }

  /* reorder xGPU and store the result, ultimately, in xGPU */
  virtual int reorderOut(CUSPARRAY *xGPU) {
    // Copy the solution, xGPU, into a temporary with the column permutation ... can't be done in place.
    thrust::copy(thrust::make_permutation_iterator(xGPU->begin(), indices->begin()),
		 thrust::make_permutation_iterator(xGPU->end(), indices->end()),
		 tempArray->begin());
    
    // Copy the temporary to the full solution.
    thrust::copy(tempArray->begin(), tempArray->end(), xGPU->begin());
    return 0;
  }


  /* Solve Analysis */
  virtual cusparseStatus_t solveAnalysis(MAT_OPERATION transA=NON_TRANSPOSE) {
    if (transA==NON_TRANSPOSE) {
      return matClass->solveAnalysis(cusparseHandle, cusparseMatDescription, 
				     cusparseSolveAnalysisInfo, transA);
    }
    else {
      if (cusparseMatDescriptionTranspose && !cusparseSolveAnalysisInfoTranspose) {
	// this version is moved into here ... because it's generated on the fly
	cusparseStatus_t stat = cusparseCreateSolveAnalysisInfo(&cusparseSolveAnalysisInfoTranspose);
	CUSPARSE_ERR_CHK(stat, "GPU_Matrix::solve()", "cusparseCreateSolveAnalysisInfo()");
	return matClass->solveAnalysis(cusparseHandle, cusparseMatDescriptionTranspose, 
				       cusparseSolveAnalysisInfoTranspose, transA);
      }
    }
    return CUSPARSE_STATUS_SUCCESS;
  }

  /* solve the triangular system */
  virtual cusparseStatus_t solve(const CUSPARRAY *bGPU, CUSPARRAY *xGPU, MAT_OPERATION transA=NON_TRANSPOSE) {
    if (transA==NON_TRANSPOSE) {
      return matClass->solve(cusparseHandle, cusparseMatDescription, 
			     cusparseSolveAnalysisInfo, bGPU, xGPU, transA);
    }
    else {
      if (cusparseMatDescriptionTranspose && cusparseSolveAnalysisInfoTranspose) {
	return matClass->solve(cusparseHandle, cusparseMatDescriptionTranspose, 
			       cusparseSolveAnalysisInfoTranspose, bGPU, xGPU, transA);
      }
    }
    return CUSPARSE_STATUS_SUCCESS;
  }

 private:
  MTYPE* matClass;
  //typename MTYPE::MatrixType mat;
  cudaStream_t stream; /*stream for running kernels that overlap with communication */
  bool streamIsACopy;

  /* These data members are for the correction step in the SpMV add */
  CUSPARRAY* tempArray;
  CUSPINTARRAYGPU* indices;

  // cusparse stuff
  cusparseHandle_t cusparseHandle;
  cusparseMatDescr_t cusparseMatDescription;
  cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfo; 

  cusparseMatDescr_t cusparseMatDescriptionTranspose;
  cusparseSolveAnalysisInfo_t cusparseSolveAnalysisInfoTranspose; 

#ifdef GPU_TIMER
  cudaEvent_t      startGPU;
  cudaEvent_t      stopGPU;
  float            dt;
  PetscInt         N; 
#endif

};

////////////////////////////////////////////////////
// Factory Class
////////////////////////////////////////////////////
class GPU_Matrix_Factory {
public:
  static GPU_Matrix_Ifc* getNew(const char * input){
    //if (type == "coo"){
    //  return new GPU_Matrix<CooMatrix<PetscInt, PetscScalar> >();
    //}
    //else 
    std::string type("csr");
    type.assign(input);
    if (type == "csr" || type=="CSR"){
      return new GPU_Matrix<CsrMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "dia" || type=="DIA"){
      return new GPU_Matrix<DiaMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "ell" || type=="ELL"){
      return new GPU_Matrix<EllMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "hyb" || type=="HYB"){
      return new GPU_Matrix<HybMatrix<PetscInt, PetscScalar> >();
    }
    else {
      // default to CSR
      return new GPU_Matrix<CsrMatrix<PetscInt, PetscScalar> >();
    }
  }
};

#endif
