#include "subscript_kernels.h"

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_SINGLE)
#define cusparseXgthr cusparseCgthr
#define cusparseXsctr cusparseCsctr
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparseXgthr cusparseZgthr
#define cusparseXsctr cusparseZsctr
#endif
#else
#if defined(PETSC_USE_REAL_SINGLE)
#define cusparseXgthr cusparseSgthr
#define cusparseXsctr cusparseSsctr
#elif defined(PETSC_USE_REAL_DOUBLE)
#define cusparseXgthr cusparseDgthr
#define cusparseXsctr cusparseDsctr
#endif
#endif


#ifndef GPU_TIMER
#define GPU_TIMER
#endif 
#undef GPU_TIMER
typedef enum {VALID_BUFFER, INVALID_BUFFER} statusFlag;

template <typename index_type, typename value_type> class GPU_Vector;

template <typename index_type, typename value_type>
class GPU_Indices {
  
 public:
  friend class GPU_Vector<index_type, value_type>;

  /////////////////////////////////////////////  
  // Constructor
  /////////////////////////////////////////////
  GPU_Indices() {
    indicesCPU = new CUSPINTARRAYCPU;
    indicesGPU = new CUSPINTARRAYGPU;
    tempGPUarray = new CUSPARRAY;
    tempCPUarray = 0;
    n = 0;
  }

  /////////////////////////////////////////////  
  // Destructor 
  /////////////////////////////////////////////
  virtual ~GPU_Indices() {
    delete indicesCPU;
    delete indicesGPU;
    delete tempGPUarray;

    // CPU work vector deallocation
    cudaError_t err = cudaFreeHost(tempCPUarray);
    CUDA_ERR_CHK_NORETURN(err, "~GPU_Indices()", "cudaFreeHost()")
  }

  /////////////////////////////////////////////  
  // buildIndices
  /////////////////////////////////////////////
  virtual cudaError_t buildIndices(const index_type * indices, const index_type numElements) {

    // first sort and test contiguity
    n = numElements;
    if (n == 0) {
      return cudaSuccess;
    }
    std::vector<index_type> Vector(indices, indices+n);
    std::sort(Vector.begin(), Vector.begin()+n);
    indicesAreContiguous = true;
    for (index_type i=0; i<n-1; ++i)
      if (Vector[i+1]!=Vector[i]+1) {
	indicesAreContiguous = false;
	break;
      }
    
    //  printf("rank %d, step 1 : sendMin=%d\tsendMax=%d\tn=%d\n",rank,sendVector.front(),sendVector.back(),n);
#if 1
    if (~indicesAreContiguous) {
      index_type min=0, max=0;
      min = *std::min_element(indices, indices+n);
      max = *std::max_element(indices, indices+n);
      n = max-min+1;
      //printf("rank %d, step 2 : min=%d\tmax=%d\tn=%d\n",rank,min,max,max-min+1);
      Vector.resize(n);
      for (index_type i=0; i<n; ++i)
	Vector[i] = min+i;
      indicesCPU->assign(&Vector[0],&Vector[0]+n);
      indicesGPU->assign(&Vector[0],&Vector[0]+n);
      tempGPUarray->resize(n);
      indicesAreContiguous=true;
    }
    else {
      indicesCPU->assign(indices,indices+n);
      indicesGPU->assign(indices,indices+n);
      tempGPUarray->resize(n);  
    }
#else
    indicesCPU->assign(indices,indices+n);
    indicesGPU->assign(indices,indices+n);
    tempGPUarray->resize(n);  
#endif
    
    // CPU work vector allocation
    cudaError_t err = cudaMallocHost((void **) &tempCPUarray, n*sizeof(value_type));
    CUDA_ERR_CHK(err, "GPU_Indices::buildIndices()", "cudaMallocHost()")
    valid_GPU_array = INVALID_BUFFER;
    return cudaSuccess;
  }
  
  /////////////////////////////////////////////  
  // resetFlag
  /////////////////////////////////////////////
  virtual void resetStatusFlag() {
    valid_GPU_array = INVALID_BUFFER;
  }

  /////////////////////////////////////////////  
  // acccesors
  /////////////////////////////////////////////
  
 private:
  
  /* CPU indices */
  CUSPINTARRAYCPU * indicesCPU;
  /* GPU indices */
  CUSPINTARRAYGPU * indicesGPU;

  /* number of indices */
  index_type n;

  /* this is a work vector for copying noncontiguous data to continguous buffers and vice versa */
  CUSPARRAY * tempGPUarray;

  /* this is a CPU work vector for copying noncontiguous data to continguous buffers and vice versa */
  value_type * tempCPUarray; 

  /* indicates where the most recently modified vector data is */
  statusFlag valid_GPU_array;    

  /* indicates whether indices are contiguous or not */
  bool indicesAreContiguous;
};

template <typename index_type, typename value_type>
class GPU_Vector {
 public:

  /////////////////////////////////////////////  
  // Constructor
  /////////////////////////////////////////////
  GPU_Vector(CUSPARRAY * gpuVector, size_t theRank) {
    /* arrays, CPU and GPU */
    cusparseStatus_t status = cusparseCreate(&cusparseHandle);
    CUSPARSE_ERR_CHK_NORETURN(status, "GPU_Vector()", "cusparseCreate");

    // This will be allocated subsequently in another
    // function reallocateHostMemory via cudaMallocHost
    CPUvector = 0;

    // At the moment the GPU vector created/allocated 
    // somewhere else. Thus, we do not delete it
    // in the destructor
    GPUvector = gpuVector;
    n = gpuVector->size();

    /* stream stuff */
    stream= 0;
    ownsPtr = false;

    rank=theRank;

#ifdef GPU_TIMER
    /* timer stuff */
    startGPU=0;
    stopGPU=0;
    
    dtSubscriptLHS=0.f;
    dtSubscriptRHS=0.f;
    dtFrom=0.f;
    dtTo=0.f;
    NSubscriptLHS=0;
    NSubscriptRHS=0;
    NFrom=0;
    NTo=0;
#endif
  }

  /////////////////////////////////////////////  
  // Destructor 
  /////////////////////////////////////////////
  virtual ~GPU_Vector() {
    cudaError_t err;
    if (cusparseHandle) {
      cusparseStatus_t status = cusparseDestroy(cusparseHandle);
      CUSPARSE_ERR_CHK_NORETURN(status, "~GPU_Vector()", "cusparseCreate");
    }

    // Kill the stream
    if (stream) {
      err = cudaStreamDestroy(stream);
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Vector()", "cudaStreamDestroy()")
      stream=0;
    }

    // delete the CPU ptr, if owned.
    if (ownsPtr) {
      err = cudaFreeHost(CPUvector); 
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Vector()", "cudaFreeHost()")
    }

#ifdef GPU_TIMER
    printf("Mean dtFrom = %1.5f\t NFrom=%d\n",dtFrom/NFrom,NFrom);
    printf("Mean dtTo   = %1.5f\t NTo=%d\n",dtTo/NTo,NTo);
    printf("Mean dtSubscriptRHS   = %1.5f\t NSubscriptRHS=%d\n",dtSubscriptRHS/NSubscriptRHS,NSubscriptRHS);
    printf("Mean dtSubscriptLHS   = %1.5f\t NSubscriptLHS=%d\n",dtSubscriptLHS/NSubscriptLHS,NSubscriptLHS);
    
    // Kill the events
    if (startGPU) {
      err = cudaEventDestroy(startGPU);
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Vector()", "cudaEventDestroy()")
      startGPU=0;
    }
    if (stopGPU) {
      err = cudaEventDestroy(stopGPU);
      CUDA_ERR_CHK_NORETURN(err, "~GPU_Vector()", "cudaEventDestroy()")
      stopGPU=0;
    }
    
    // reset vars
    dtSubscriptRHS=0.f;
    dtSubscriptLHS=0.f;
    dtFrom=0.f;
    dtTo=0.f;
    NSubscriptRHS=0;
    NSubscriptLHS=0;
    NFrom=0;
    NTo=0;
#endif
  }


  /////////////////////////////////////////////
  // allocate host memory as page locked 
  /////////////////////////////////////////////
  virtual cudaError_t allocateHostMemory() {
    cudaError_t err = cudaMallocHost((void **) &CPUvector, n*sizeof(value_type));
    CUDA_ERR_CHK(err, "GPU_Vector::allocateHostMemory()", "cudaMallocHost()")
    ownsPtr=true;
    return err;
  }
  virtual value_type * getHostMemoryPtr() {
    return CPUvector;
  }

  virtual void assignHostMemory(value_type * hostVector) {
    ownsPtr=false;
    CPUvector = hostVector;
  }
  
  /* build streams and events */
  virtual cudaError_t buildStreamsAndEvents() {
    
    cudaError_t err=cudaSuccess;
    if (!stream) {
      err = cudaStreamCreate(&stream);
      CUDA_ERR_CHK(err, "GPU_Vector::buildStreamsAndEvents()", "cudaStreamCreate()")
    }

#ifdef GPU_TIMER
    err = cudaEventCreate(&startGPU);
    CUDA_ERR_CHK(err, "GPU_Vector::buildStreamsAndEvents()", "cudaEventCreate()")
    err = cudaEventCreate(&stopGPU);
    CUDA_ERR_CHK(err, "GPU_Vector::buildStreamsAndEvents()", "cudaEventCreate()")
#endif
    return err;
  };

  //////////////////////////////
  // CopyToGPUAll
  //////////////////////////////
  virtual cudaError_t copyToGPUAll() {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif    
    value_type * gpuPtr = thrust::raw_pointer_cast(&(GPUvector->begin())[0]);
    //printf("%d : copyToGPUAll for %p\n",rank,CPUvector);
    //for(size_t i=0; i<n; ++i)
    // printf("%d : %1.5g\n",rank,CPUvector[i]);
    cudaError_t err=cudaSuccess;
    err = cudaMemcpyAsync(gpuPtr, 
			  CPUvector, n*sizeof(value_type), 
			  cudaMemcpyHostToDevice, stream);
    CUDA_ERR_CHK(err, "GPU_Vector::copyToGPUAll()", "cudaMemcpyAsync()")

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtTo+=dt;
    NTo++;
#endif

    //Have to synchronize the stream 
    err = cudaStreamSynchronize(stream);
    CUDA_ERR_CHK(err, "GPU_Vector::copyToGPUAll()", "cudaStreamSynchronize()")
    return err;    
  }



  //////////////////////////////
  // CopyToGPUSome
  //////////////////////////////
  virtual int copyToGPUSome(CUSPARRAY* varray,
			    GPU_Indices<index_type, value_type>* indices) {

    value_type *tempCPUarray=indices->tempCPUarray;
    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    CUSPINTARRAYCPU *indicesCPU=indices->indicesCPU;
    index_type m = indicesGPU->size();

    /* return if size is zero */
    if (m == 0) {
      return cudaSuccess;
    }
    // Going to do this in 3 steps
    // 1) Move Scattered array to contiguous array on the host
    // 2) Move contiguous array to device via cudaMemcpyAsync()
    // 3) Move contiguous device data to scattered data on the device.
    //if (ownsPtr) printf("copyToGPUSome\n");

    //std::cout << "copyToGPUSome stream=\n"<<stream <<std::endl;

    // Step 1: copy into correct locations
    index_type * cpuInd = thrust::raw_pointer_cast(&(indicesCPU->begin())[0]);
    if (!indices->indicesAreContiguous) {
      for (index_type i=0; i<m; ++i)
	tempCPUarray[i] = CPUvector[cpuInd[i]];
    }
    //printf("%d : copyToGPUSome for %p\n",rank,CPUvector);
    //for(size_t i=0; i<m; ++i)
    // printf("%d : (%d, %1.5g)\n",rank,tempCPUarray[i],cpuInd[i]);

    // Step 2: asynchronous memory transfer to the device
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    value_type * gpuPtr;
    value_type * cpuPtr;
    if (!indices->indicesAreContiguous) {
      gpuPtr = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
      cpuPtr = tempCPUarray;
    }
    else {
      gpuPtr = thrust::raw_pointer_cast(&(varray->begin())[0])+cpuInd[0];
      cpuPtr = CPUvector + cpuInd[0];
    }        
    cudaError_t err=cudaSuccess;
    err = cudaMemcpyAsync(gpuPtr, cpuPtr, m*sizeof(value_type), 
			  cudaMemcpyHostToDevice, stream); 
    CUDA_ERR_CHK(err, "GPU_Vector::copyToGPUSome()", "cudaMemcpyAsync()")
#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtTo+=dt;
    NTo++;
#endif
    indices->valid_GPU_array = VALID_BUFFER;
    // Step 3: this function scatters data from the contiguous buffer to the correct locations in the vector
    // on the device.
    if (!indices->indicesAreContiguous) {
      err = copySomeFromContiguousBuffer(varray, indices);
      if (err!=cudaSuccess) return err;
    }
    //Have to synchronize the stream 
    //err = cudaThreadSynchronize();
    err = cudaStreamSynchronize(stream);
    CUDA_ERR_CHK(err, "GPU_Vector::copyToGPUSome()", "cudaStreamSynchronize()")
    return err;
  }

  //////////////////////////////
  // CopyFromGPUAll
  //////////////////////////////
  virtual int copyFromGPUAll() {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    value_type * gpuPtr = thrust::raw_pointer_cast(&(GPUvector->begin())[0]);
    cudaError_t err=cudaSuccess;
    err = cudaMemcpyAsync(CPUvector, gpuPtr, 
			  n*sizeof(value_type), cudaMemcpyDeviceToHost, stream);
    CUDA_ERR_CHK(err, "GPU_Vector::copyToGPUAll()", "cudaMemcpyAsync()")
    
    //Have to synchronize the stream before the MPI buffers are filled.
    err = cudaStreamSynchronize(stream);
    CUDA_ERR_CHK(err, "GPU_Vector::copyFromGPUAll()", "cudaStreamSynchronize()")

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtFrom+=dt;
    NFrom++;      
#endif
    return err;    
  }
  
  
  //////////////////////////////
  // copyFromGPUSome
  //////////////////////////////
  virtual cudaError_t copyFromGPUSome(CUSPARRAY* varray,
			      GPU_Indices<index_type, value_type>* indices) {

    value_type *tempCPUarray=indices->tempCPUarray;
    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    CUSPINTARRAYCPU *indicesCPU=indices->indicesCPU;
    index_type m = indicesGPU->size();
    // Going to do this in 3 steps
    // 1) Move Scattered array to contiguous array on the device ... this is done in VecCUSPCopySomeToContiguousBufferGPU
    // 2) Move contiguous array to host via cudaMemcpyAsync()
    // 3) On the CPU, move data in contiguous buffer to the correct scattered locations.

    // Step 1:
    cudaError_t err=cudaSuccess;
    if (!indices->indicesAreContiguous) {
      err = copySomeToContiguousBuffer(varray, indices);
      if (err!=cudaSuccess) return err;
    }
    // Step 2:
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    value_type * gpuPtr;
    value_type * cpuPtr;
    index_type * cpuInd = thrust::raw_pointer_cast(&(indicesCPU->begin())[0]);
    if (!indices->indicesAreContiguous) {
      gpuPtr = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
      cpuPtr = tempCPUarray;
    }
    else {
      gpuPtr = thrust::raw_pointer_cast(&(varray->begin())[0])+cpuInd[0];
      cpuPtr = CPUvector+cpuInd[0];
    }
    err = cudaMemcpyAsync(cpuPtr, gpuPtr, m*sizeof(value_type), 
			  cudaMemcpyDeviceToHost, stream); 
    CUDA_ERR_CHK(err, "GPU_Vector::copyFromGPUSome()", "cudaMemcpyAsync()")
    
    //Have to synchronize the stream 
    err = cudaStreamSynchronize(stream);
    CUDA_ERR_CHK(err, "GPU_Vector::copyFromGPUSome()", "cudaStreamSynchronize()")

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtFrom+=dt;
    NFrom++;
#endif

    //printf("%d : copyFromGPUSome for %p\n",rank, CPUvector);
    //for(size_t i=0; i<m; ++i)
    // printf("%d : (%d, %1.5g)\n",rank,tempCPUarray[i],cpuInd[i]);
    // Step 3: now copy into correct locations
    if (!indices->indicesAreContiguous) {
      for (index_type i=0; i<m; ++i)
	CPUvector[cpuInd[i]] = tempCPUarray[i];
    }
    return err;
  }


  //////////////////////////////
  // copySomeToContiguousBuffer
  //////////////////////////////
  virtual cudaError_t copySomeToContiguousBuffer(CUSPARRAY* varray,
						 GPU_Indices<index_type, value_type>* indices) {

    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    index_type m = indicesGPU->size();
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    cudaError_t err=cudaSuccess;
    if (indices->valid_GPU_array==INVALID_BUFFER && !indices->indicesAreContiguous) {
      index_type * indPtr = thrust::raw_pointer_cast(&(indicesGPU->begin())[0]);
      value_type * gpuPtrLHS = thrust::raw_pointer_cast(&(varray->begin())[0]);
      value_type * gpuPtrRHS = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
#if 1
      err = subscriptRHS(stream, m, indPtr, gpuPtrLHS, gpuPtrRHS);
      CUDA_ERR_CHK(err, "GPU_Vector::copySomeToContiguousBuffer()", "subscriptRHS");
#else
      err = (cudaError_t) cusparseSetStream(cusparseHandle, stream);
      CUDA_ERR_CHK(err, "GPU_Vector::copySomeToContiguousBuffer()", "cusparseSetStream");
      err = (cudaError_t) cusparseXgthr(cusparseHandle, m, gpuPtrLHS, gpuPtrRHS, indPtr, CUSPARSE_INDEX_BASE_ZERO); 
      CUDA_ERR_CHK(err, "GPU_Vector::copySomeToContiguousBuffer()", "cusparseXgthr");
#endif 
    }
    indices->valid_GPU_array=VALID_BUFFER;

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtSubscriptRHS+=dt;
    NSubscriptRHS++;
#endif
    return err;
  }


  ////////////////////////////////////
  // copySomeFromContiguousBuffer
  ////////////////////////////////////
  virtual cudaError_t copySomeFromContiguousBuffer(CUSPARRAY* varray,
						   GPU_Indices<index_type, value_type>* indices) {

    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    index_type m = indicesGPU->size();
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    cudaError_t err=cudaSuccess;
    if (indices->valid_GPU_array==VALID_BUFFER) {
      index_type * indPtr = thrust::raw_pointer_cast(&(indicesGPU->begin())[0]);
      value_type * gpuPtrRHS = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
      value_type * gpuPtrLHS = thrust::raw_pointer_cast(&(varray->begin())[0]);
#if 1
      err = subscriptLHS(stream, m, indPtr, gpuPtrRHS, gpuPtrLHS);
      CUDA_ERR_CHK(err, "GPU_Vector::copySomeFromContiguousBuffer()", "subscriptLHS");
#else
      err = (cudaError_t) cusparseSetStream(cusparseHandle, stream);
      CUDA_ERR_CHK(err, "GPU_Vector::copySomeFromContiguousBuffer()", "cusparseSetStream");
      err = (cudaError_t) cusparseXsctr(cusparseHandle, m, gpuPtrRHS, indPtr, gpuPtrLHS, CUSPARSE_INDEX_BASE_ZERO); 
      CUDA_ERR_CHK(err, "GPU_Vector::copySomeFromContiguousBuffer()", "cusparseXsctr")
#endif 
    }

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);
    dtSubscriptLHS+=dt;
    NSubscriptLHS++;
#endif
    return err;
  }

private:
  cusparseHandle_t cusparseHandle; /* Handle to cusparse */
  cudaStream_t     stream;    /* it's easier to put it in here */
  value_type *   CPUvector;  /* pointer to page-locked host memory */
  CUSPARRAY*     GPUvector;  /* this always holds the ptr to GPU data */
  index_type     n; /* size of the vector */
  bool           ownsPtr;
  size_t         rank;

#ifdef GPU_TIMER
  cudaEvent_t      startGPU;
  cudaEvent_t      stopGPU;
  float      dtSubscriptLHS;
  float      dtSubscriptRHS;
  float      dtFrom;
  float      dtTo;
  index_type NSubscriptLHS; /* number of times that VecCUSPCopyFromGPU has been called */
  index_type NSubscriptRHS; /* number of times that VecCUSPCopyFromGPU has been called */
  index_type NFrom; /* number of times that VecCUSPCopyFromGPU has been called */
  index_type NTo; /* number of times that VecCUSPCopyToGPU has been called */
#endif
};
